/*
 * clock for root window.
 */

/*
 * - finish alarms set up
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <locale.h>
#include <string.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <librootapp/librootapp.h>
#include "config.h"

#define EVENTMASK VisibilityChangeMask | ExposureMask

#define pi 3.1415927

#define DIGITAL 1
#define ANALOG  2
#define BINARY  3

#define ROUND 0
#define RECT  1


/* Global varriables */
static int end_program  = FALSE;
static int display_mode = DIGITAL;

static char       timestr[MAXSTRLEN + 1];
static char       datestr[MAXSTRLEN + 1];
static char      *dpyname    = NULL;
static char      *fontname   = NULL;
static char      *geometry   = NULL;
static char      *dgeometry  = NULL;
static char      *mcolorname = NULL;
static char      *scolorname = NULL;
static StrLst    *alarms     = NULL;
static char      *alarm_cmd  = NULL;
static char      *tformat    = NULL;
static char      *dformat    = NULL;
static int        aradius    = 10;
static int        bradius    = 5;
static int        draw       = ROUND;
static int        h, m, s;
static int        wd, md, mn;
static unsigned long cm = 0, cs = 0;



/* Command line options and config file entries */
#define OHLP(h)          {'h', NULL, NULL, O_HELP,   NULL, FALSE, h}
#define OVRS(h)          {'v', NULL, NULL, O_VERS,   NULL, FALSE, h}
#define OINT(k, o, v, h) {k,   o,    NULL, C_INT,    v,    FALSE, h}
#define OSTR(k, o, v, h) {k,   o,    NULL, C_STRING, v,    FALSE, h}
#define OBOL(k, o, v, h) {k,   o,    NULL, C_BOOL,   v,    FALSE, h}
#define OSTL(k, o, v, h) {k,   o,    NULL, C_STRLST, v,    FALSE, h}
#define OSTP             {0,   NULL, NULL, C_NONE,   NULL, FALSE, NULL}
static Option opttab[] = {
    OHLP("display help"),
    OVRS("display version"),
    OSTR('d', "display",  &dpyname,      "display name"),
    OINT('m', "mode",     &display_mode, "display mode"),
    OSTR('c', "color",    &mcolorname,   "main color (highlight)"),
    OSTR('s', "color",    &scolorname,   "secondary color (borders)"),
    OSTR('f', "font",     &fontname,     "font to use"),
    OSTR('g', "geom",     &geometry,     "time geometry"),
    OSTR('z', "dategeom", &dgeometry,    "date geometry"),
    OINT('a', "a-radius", &aradius,      "analog clock radius"),
    OINT('b', "b-radius", &bradius,      "binary clock radius (dots)"),
    OSTR('t', "timefmt",  &tformat,      "time format"),
    OSTR('u', "datefmt",  &dformat,      "date format"),
    OBOL('r', NULL,    &draw, "draw rectangles (not rounds) for binary clock"),
    OSTL('l', "alarm",    &alarms,       "alarm time"),
    OSTR('x', "cmd",      &alarm_cmd,    "alarm play command"),
    OSTP
};



/****** UPDATING FUNCTIONS ******/

int update_now() {
    time_t     tnow;
    struct tm *now;
    int        ret     = 0;
    char      *timefmt = tformat ? tformat : "%T";
    char      *datefmt = dformat ? dformat : "%A %d %B";

    time(&tnow);
    now = localtime(&tnow);
    h = now->tm_hour;
    m = now->tm_min;
    s = now->tm_sec;
    wd = now->tm_wday;
    if (wd == 0) wd = 7;
    md = now->tm_mday;
    mn = now->tm_mon + 1;
    bzero(timestr, MAXSTRLEN + 1);
    bzero(datestr, MAXSTRLEN + 1);
    strftime(timestr, MAXSTRLEN, timefmt, now);
    strftime(datestr, MAXSTRLEN, datefmt, now);
    if ((now->tm_hour == 0) && (now->tm_min ==0) && (now->tm_sec == 0))
        ret = 1;
    if ((now->tm_sec == 0) && alarm_cmd && alarms) {
        char    thistime[MAXSTRLEN + 1];
        StrLst *alarm   = alarms;
        int     execute = FALSE;

        strftime(thistime, MAXSTRLEN, "%H:%M", now);
        while (alarm) {
            if (strcmp(thistime, alarm->entry) == 0) execute = TRUE;
            alarm = alarm->next;
        }
        if (execute) my_system(alarm_cmd, NULL);
    }
    return ret;
}


/****** DISPLAYING FUNCTIONS ******/

void calc_point(int position, Bool is_hour, int length, int *x, int *y) {
    int a;

    if (is_hour)
        a = 30 * (position - 3);
    else
        a = 6 * (position - 15);

    *x = aradius + length * cos(a*pi/180);
    *y = aradius + length * sin(a*pi/180);
}


void set_default_segments(XSegment segments[]) {
    int i;

    /* All */
    for (i = 0 ; i < 12 ; i++) {
        int x, y;
        calc_point(i, True, aradius, &x, &y);
        segments[i].x1 = x;
        segments[i].y1 = y;
        segments[i].x2 = x;
        segments[i].y2 = y;
    }
}


void draw_analogtime(RootApp *app, XSegment segments[]) {
    int x, y;

    /* Hours */
    calc_point(h, True, aradius / 2, &x, &y);
    segments[12].x1 = aradius;
    segments[12].y1 = aradius;
    segments[12].x2 = x;
    segments[12].y2 = y;

    /* Minutes */
    calc_point(m, False, 2 * aradius / 3, &x, &y);
    segments[13].x1 = aradius;
    segments[13].y1 = aradius;
    segments[13].x2 = x;
    segments[13].y2 = y;

    set_color(cm);
    /* Markers, hours' hand, minutes' hand */
    XDrawSegments(display, app-> bgpixmap, gc, segments, 14);

    /* seconds */
    set_color(cs);
    calc_point(s, False, 3 * aradius / 4, &x, &y);
    XDrawLine(display, app-> bgpixmap, gc, aradius, aradius, x, y);
}


void get_bits(int t, int b[]) {
    int v = t, i;

    for (i = 0 ; i < 4 ; i++) b[i] = 0;
    i = 0;
    while (v != 0) {
        if (i == 4) {
            fprintf(stderr, "Error : number is too big for 4 bits (%d)\n", t);
            exit(1);
        }
        b[i] = v % 2;
        v = v / 2;
        i++;
    }
}


void draw_round_bits(RootApp *app, int t, int r, int p) {
    int i = 0;
    int b[4];

    get_bits(t, b);
    for (i = 0 ; i < 4 ; i++) {
        int x = p;
        int y = 3 * (r + r / 2) - i * (r + r / 2);
        if (b[i]) {
            set_color(cm);
            XDrawArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
            XFillArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
        } else
            set_color(cs);
            XDrawArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
    }
}


void draw_rect_bits(RootApp *app, int t, int w, int p) {
    int i = 0;
    int b[4];

    get_bits(t, b);
    for (i = 0 ; i < 4 ; i++) {
        int x = p;
        int y = 3 * w - i * w;
        if (b[i]) {
            set_color(cm);
            XDrawRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
            XFillRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
        } else
            set_color(cs);
            XDrawRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
    }
}


void draw_binarytime(RootApp *app) {
    switch (draw) {
        case ROUND:
            draw_round_bits(app, h / 10, bradius, 0);
            draw_round_bits(app, h % 10, bradius, bradius + bradius / 2);

            draw_round_bits(app, m / 10, bradius, bradius * 4);
            draw_round_bits(app, m % 10, bradius, bradius * 5 + bradius / 2);

            draw_round_bits(app, s / 10, bradius, bradius * 8);
            draw_round_bits(app, s % 10, bradius, bradius * 9 + bradius / 2);
            break;
        case RECT:
            draw_rect_bits(app, h / 10, bradius, 0);
            draw_rect_bits(app, h % 10, bradius, bradius + bradius / 2);

            draw_rect_bits(app, m / 10, bradius, bradius * 4);
            draw_rect_bits(app, m % 10, bradius, bradius * 5 + bradius / 2);

            draw_rect_bits(app, s / 10, bradius, bradius * 8);
            draw_rect_bits(app, s % 10, bradius, bradius * 9 + bradius / 2);
            break;
    }
}


void draw_binarydate(RootApp *app) {
    switch (draw) {
        case ROUND:
            draw_round_bits(app, wd, bradius, 0);

            draw_round_bits(app, md / 10, bradius, bradius * 2 + bradius / 2);
            draw_round_bits(app, md % 10, bradius, bradius * 4);

            draw_round_bits(app, mn, bradius, bradius * 6 + bradius / 2);
            break;
        case RECT:
            draw_rect_bits(app, wd, bradius, 0);

            draw_rect_bits(app, md / 10, bradius, bradius * 2 + bradius / 2);
            draw_rect_bits(app, md % 10, bradius, bradius * 4);

            draw_rect_bits(app, mn, bradius, bradius * 6 + bradius / 2);
            break;
    }
}


void draw_time(RootApp *app, XSegment segs[]) {
    if (app) {
        clear_bg(app);
        set_color(cm);
        switch (display_mode) {
            case DIGITAL: root_print(app, timestr, 0, 1); break;
            case ANALOG:  draw_analogtime(app, segs);     break;
            case BINARY:  draw_binarytime(app);           break;
        }
        display_app(app);
    }
}


void draw_date(RootApp *app) {
    if (app) {
        clear_bg(app);
        set_color(cm);
        switch (display_mode) {
            case DIGITAL: root_print(app, datestr, 0, 0); break;
            case ANALOG:  root_print(app, datestr, 0, 0); break;
            case BINARY:  draw_binarydate(app);           break;
        }
        display_app(app);
    }
}


/****** PROCESSING FUNCTIONS ******/

void resize_time(RootApp *app) {
    int width, height;

    switch (display_mode) {
        case DIGITAL:
            set_to_text_size(app, geometry, timestr);
            break;
        case ANALOG:
            resize(app, geometry, 0, 0, 2 * aradius + 1, 2 * aradius + 1);
            break;
        case BINARY:
            width = bradius * 10 + bradius / 2 + 1;
            switch (draw) {
                case RECT: height = 4 * bradius; break;
                case ROUND:
                default: height = 4 * (bradius + bradius / 2); break;
            }
            resize(app, geometry, 0, 0, width, height);
            break;
    }
}


void resize_date(RootApp *app) {
    int width, height;

    switch (display_mode) {
        case DIGITAL:
        case ANALOG:
            set_to_text_size(app, dgeometry, datestr);
            break;
        case BINARY:
            width = bradius * 7 + bradius / 2 + 1;
            switch (draw) {
                case RECT: height = 4 * bradius; break;
                case ROUND:
                default: height = 4 * (bradius + bradius / 2); break;
            }
            resize(app, dgeometry, 0, 0, width, height);
            break;
    }
}


void resize_windows(RootApp *t, RootApp *d) {
    resize_time(t);
    resize_date(d);
}


void main_loop(RootApp *t, RootApp *d) {
    XSegment segments[14];
    XEvent   event;
    int      cont = TRUE;
    int      mode = display_mode;

    set_default_segments(segments);
    draw_time(t, segments);
    draw_date(d);
    while (cont) {
        nextevent_or_timeout(&event, 1000);

        if ((update_now() == 1) && d && (display_mode != BINARY))
            set_to_text_size(d, dgeometry, datestr);

        if (mode != display_mode) {
            resize_windows(t, d);
            mode = display_mode;
        }

        draw_time(t, segments);
        draw_date(d);

        if (end_program) cont = FALSE;
    }
}


/****** SIGNALS ******/

void install_handler();


void signals_handler(int sig) {
    switch (sig) {
        case SIGUSR1:
            display_mode++;
            if (display_mode > BINARY) display_mode = DIGITAL;
            break;
        case SIGUSR2:
            display_mode--;
            if (display_mode < DIGITAL) display_mode = BINARY;
            break;
        case SIGINT:
        case SIGTERM:
        case SIGQUIT:
            end_program = TRUE;
            break;
    }
}


void install_handler() {
    signal(SIGTERM, signals_handler);
    signal(SIGQUIT, signals_handler);
    signal(SIGINT,  signals_handler);
    signal(SIGUSR1, signals_handler);
    signal(SIGUSR2, signals_handler);
}


void manage_signals() {
    struct sigaction sa;

    sa.sa_handler = SIG_IGN;
#ifdef SA_NOCLDWAIT
    sa.sa_flags = SA_NOCLDWAIT;
#else
    sa.sa_flags = 0;
#endif
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD, &sa, NULL);
    install_handler();
}


/****** MAIN ******/


int main(int argc, char *argv[]) {
    RootApp *rtime = NULL;
    RootApp *rdate = NULL;

    /* General init */
    manage_signals();
    setlocale(LC_ALL, "");
    init_rootapp(PACKAGE, PROGRAM, VERSION, AUTHORS);
    parse_argv(opttab, argc, argv);
    update_now();

    /* X11 init */
    open_x11(dpyname);
    if (fontname) load_font(fontname);
    if (mcolorname) cm = get_color(mcolorname);
    if (scolorname) cs = get_color(scolorname);

    /* Creation */
    rtime = create_app(EVENTMASK, geometry);
    resize_time(rtime);
    if (dgeometry) {
        rdate = create_app(EVENTMASK, dgeometry);
        resize_date(rdate);
    }

    main_loop(rtime, rdate);

    /* Clean */
    close_app(rtime);
    if (rdate) close_app(rdate);
    close_x11();
    free_vars(opttab);

    return 0;
}
